# $Id: Entry.pm,v 0.1 2004/05/01 00:00:00 luka Exp $
# This file is part of PerlPanel.
# 
# PerlPanel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PerlPanel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PerlPanel; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Copyright: (C) 2004 Luka Frelih <luka@ljudmila.org>
#
# Notes:
# 	- this file is at a very early stage (read: what i needed now)
#	- the public api is not determined yet
#	- reference: http://freedesktop.org/Standards/desktop-entry-spec/0.9.4/
#	- some ideas from: http://mail.gnome.org/archives/gtk-devel-list/2004-April/msg00228.html
#	- i wish for constructive comments on this & patches from everyone interested
#	- i want to bundle this with perlpanel now and to make it better and put in cpan
#	- a goal is easy handling of desktop entry files properly for reading & editing
#	- no docs yet - use the source
#
# Missing functionality: 
#	- encoding & escaping awareness
#	- value type awareness
#	- saving more than the default group
#
package Desktop::Entry;

use strict;

our $VERSION="0.1";

our @string_keys = ('Type', 'Encoding', 'TryExec', 'Exec', 'Path', 'SwallowExec', 'Dev', 'FSType', 'MountPoint', 'URL', 'StartupWMClass' );

our @string_list_keys = ('Actions','SortOrder', 'Categories', 'OnlyShowIn', 'NotShowIn');

our @localestring_keys = ('Name', 'GenericName', 'Comment', 'Icon', 'UnmountIcon', 'SwallowTitle');

our @boolean_keys = ('NoDisplay', 'Hidden', 'Terminal', 'ReadOnly', 'StartupNotify');

our @boolean_list_keys = ();

our @numeric_keys = ('Version');

our @regexp_list_keys = ('FilePattern','MimeType');

sub new {
	my $self		= {};
	$self->{package}	= shift;
	$self->{id} = shift;	
	$self->{groups} = {};
	$self->{locales} = {};
	$self->{group} = {};
	$self->{locale} = $ENV{LANG};
	$self->{interesting_locales} = {};

	bless($self, $self->{package});
	return $self;
}

sub get {
	my $self = shift;
	my $key = shift;
	my $grp = shift; #TODO: groups...

	if (!defined($self->{locale_order})) {
	
		my %lh=%{$self->{interesting_locales}};

		my @ll;

		foreach my $key (sort {$lh{$b} <=> $lh{$a	}} keys %lh){
#			print "SORT $key = $lh{$key}\n";
			push @{$self->{locale_order}},$key;
		};
	}


	if (my $key=$self->{group}->{$key}) {
		foreach my $loc (@{$self->{locale_order}}) {
			return $key->{$loc} if defined($key->{$loc});
		}
	}

	return undef;
}

sub parse_file {
	my $self=shift;
	my $filename=trim(shift);

	return 0 unless $filename;

	open(SOURCE, "< $filename");

	while (<SOURCE>) {

		$self->parse_line($_);
	}
	
	close(SOURCE);

	$self->set_group('Desktop Entry');

print "Name: ".$self->get('Name')."\n";
print "GenericName: ".$self->get('GenericName')."\n";
print "Exec: ".$self->get('Exec')."\n";
print "Icon: ".$self->get('Icon')."\n";
print "Version: ".$self->get('Version')."\n";

	return 1;	
}


sub write_file {
	my $self=shift;
	my $filename=trim(shift);

	print "Writing file $filename\n";
	return 0 unless $filename;

	open(ENTRY, "> $filename");

	$self->set_group('Desktop Entry');

	print ENTRY "[Desktop Entry]".$/;
#	print "[Desktop Entry]".$/;

	my $g=$self->{group};

	foreach my $key (keys %{$g}) {
#		print "K $key\n";

		my $k=$g->{$key};

		foreach my $loc (keys %{$k}) {
#			print "L $loc\n";

			my $locs=($loc eq 'C') ? '' : "[$loc]";

			print ENTRY $key.$locs."=".$k->{$loc}.$/;
#			print $key.$locs."=".$k->{$loc}.$/;
		}
	}
			
	close(ENTRY);

	return 1;	
}


sub parse_line {
	my $self=shift;
	my $l=trim(shift);

#	print "PL: $l - ";

	if (line_is_comment($l)) {
#		print "Comment\n";
	} elsif (line_is_group($l)) {
#		print "Group\n";
		$self->parse_group($l);	
	} elsif (line_is_entry($l)) {
#		print "Entry\n";
		$self->parse_entry($l);
	}

}

sub line_is_comment {
	my $l=shift;

	return $l =~ /^\#/ if $l;
	return 0;
}

sub line_is_group {
	my $l=shift;

	return $l =~ /^\[.+\]$/ if $l;
	return 0;

}

sub parse_group {
	my $self=shift;
	my $l=shift;

	if ($l =~ /^\[(.+)\]$/) {
		$self->{group} = $self->{groups}->{$1} = {};
		print "Group: $1\n";
	}
}

sub line_is_entry {
	my $l=shift;

	return $l =~ /\S+\s*=\s*\S+/ if $l;
	return 0;

}

sub parse_entry {
	my $self=shift;
	my $l=shift;

	if ($l =~ /^(.+?)\s*=\s*(.+)$/) {
		my $key = $1;
		my $val = $2;
		my $loc = 'C';

		if ($key =~ s/\[(.+?)\]$//) {
			$loc = $1;
		}
#		print "K: $key\tL:$loc\tV: $val\n";
	
		$self->{group}->{$key}->{$loc} = $val;

		my @ltoks=locale_tokens($loc);

#		print "LM: ".locale_match($loc,$self->{locale})."\n";
	
		if (!defined($self->{locales}->{$loc})) {
			my $m;
			if ($m=locale_match($loc,$self->{locale})) {
				$self->{interesting_locales}->{$loc}=$m;
			}
			$self->{locales}->{$loc}=$m;
		}
	}
}

sub set_group {
	my $self=shift;
	my $grp=shift;

	if (defined($self->{groups}->{$grp})) {
		$self->{group} = $self->{groups}->{$grp};
	}

}

#
# tokens
#

#
# this returns a positive float if candidate is same or more general version of test
# 
sub locale_match {
	my $candidate = shift;	# these are refs to token lists
	my $test = shift;	

	return 1 if ($candidate eq $test);
	return 0.1 if ($candidate eq "C");
	
	my @c=locale_tokens($candidate);
	my @t=locale_tokens($test);
	my $m;
	
	if (($c[0]) && ($c[1]) && ($c[2]) && ($c[3])) {
		return 0;
	}

	my $tl;
	if ($c[0]) {
		$tl = ($c[0] eq $t[0]);
		return 0 unless $tl;
		$m += 0.25;
	}

	my $tc;
	if ($c[1]) {
		$tc = ($c[1] eq $t[1]);
		return 0 unless $tc;
		$m += 0.25;
	}

	my $te;
	if ($c[2]) {
		$te = ($c[2] eq $t[2]);
		return 0 unless $te;
		$m += 0.25;
	}

	my $tm;
	if ($c[3]) {	
		$tl = ($c[3] eq $t[3]);
		return 0 unless $tl;
		$m += 0.25;
	}

	#hopefully?
	return $m;	

}

sub locale_tokens {
	my $s=shift;
	my $lang;
	my $country;
	my $encoding;
	my $modifier;

	if ($s =~ /^(.+?)($|[_.@])/) { 		#this should always suceed
		$lang=$1;
	}

	if ($s =~ /_(.+?)($|[.@])/) {
		$country=$1;
	}

	if ($s =~ /\.(.+?)($|[@])/) {
		$encoding=$1;
	}

	if ($s =~ /@(\w+?)$/) {
		$modifier=$1;
	}

#	print "L:$lang C:$country E:$encoding M:$modifier\n"; 
	return ($lang,$country,$encoding,$modifier);
}


#
# utils
#

sub dump {
use Data::Dumper;

	my $self=shift;

	print Dumper($self);
}

sub trim {
	my $s=shift;

	$s=~s/^\s*(.*?)\s*$/\1/;

	return $s;
}

sub get_exec {
	my $self=shift;
	my $conf=shift;

	my $s=$self->get('Exec');

#	print "IN: $s ";

	$s =~ s/%f/$conf->{file}/g;
	$s =~ s/%F/$conf->{files}/g;
	$s =~ s/%u/$conf->{url}/g;
	$s =~ s/%U/$conf->{urls}/g;
	$s =~ s/%d/$conf->{dir}/g;
	$s =~ s/%D/$conf->{dirs}/g;
	$s =~ s/%d/$conf->{dir}/g;
	$s =~ s/%n/$conf->{filename}/g;
	$s =~ s/%N/$conf->{filenames}/g;
	$s =~ s/%i/$conf->{icon}/g;
	$s =~ s/%c/$conf->{name}/g;
	$s =~ s/%k/$conf->{entryfile}/g;
	$s =~ s/%v/$conf->{device}/g;
	$s =~ s/%m/$conf->{icon}/g; #deprecated?

#	print "OUT: $s\n";

	return $s;

}

sub parse_value {
# The value types recognized are string, localestring, regexp, boolean (encoded as the string true/false), and numeric.
# The difference between string and localestring is that the value for a string key must contain only ASCII characters and while the value of a localestring key may contain UTF-8 characters. (See section 5.)
# Some keys can have multiple values; these should be separated by a semicolon. Those keys which have several values should have a semicolon as the trailing character. For lists of strings, semicolons are simply not allowed in the strings, there is no escape mechanism. 


}

sub value_unescape {
# The escape sequences \s, \n, \t, \r, and \\ are supported, meaning ASCII space, newline, tab, carriage return, and backslash, respectively.

	my $s = shift;

#TODO

	return $s;

}

#my $de=new Desktop::Entry('konsole');
#$de->parse_line("     #[Desktop Entry]     ");
#$de->parse_line("     [Desktop Entry]     ");
#$de->parse_line("     key=value     ");
#$de->parse_line("     key[sl_SI]=value1     ");
#$de->parse_line("     key[sl_SI.UTF-8]=value2     ");
#$de->parse_line("     key[sl]=value3     ");
#$de->parse_line("     key[sl_SI.UTF-8\@mod]=value4     ");
#$de->parse_line("     key[en_US]=value5     ");
#$de->parse_line("     key[de]=value6     ");
#$de->parse_line("     key[fr]=value7     ");

#while (<>) {
#	$de->parse_line($_);
#}


#$de->set_group('Desktop Entry');

#print "Name: ".$de->get('Name')."\n";
#print "GenericName: ".$de->get('GenericName')."\n";
#print "Exec: ".$de->get('Exec')."\n";
#print "Icon: ".$de->get('Icon')."\n";
#print "Version: ".$de->get('Version')."\n";

#$de->dump;

1;


=pod

Type	 There are 4 types of desktop entries: Application(1), Link(2), FSDevice(3) and Directory(4). 	
string	YES	YES

Version	Version of Desktop Entry Specification (While the version field is not required to be present, it should be in all newer implementations of the Desktop Entry specification. If the version number is not present, a "pre-standard" desktop entry file is to be assumed). 	
numeric	NO	YES	1-4

Encoding	Encoding of the whole desktop entry file (UTF-8 or LegacyMixed). 	
string	YES	YES	1-4

Name	Specific name of the application, for example "Mozilla". 	
localestring	YES	YES	1-4

GenericName	Generic name of the application, for example "Web Browser". 	
localestring	NO	YES	1-4

NoDisplay	NoDisplay means "this application exists, but don't display it in the menus". This can be useful to e.g. associate this application with mimetypes, so that it gets launched from a file manager (or other apps), without having a menu entry for it (there are tons of good reasons for this, including e.g. the netscape -remote, or kfmclient openURL kind of stuff). 	
boolean	NO	NO	1-4

Comment	Tooltip for the entry, for example "View sites on the Internet"; should not be redundant with Name or GenericName. 	
localestring	NO	YES	1-4

Icon	Icon to display in file manager, menus, etc. If the name is an absolute path, the given file will be used. If the name is not an absolute path, an implementation-dependent search algorithm will be used to locate the icon. Icons may be localized with the Icon[xx]= syntax. 	
string	NO	YES	1-4

Hidden	Hidden should have been called Deleted. It means the user deleted (at his level) something that was present (at an upper level, e.g. in the system dirs). It's strictly equivalent to the .desktop file not existing at all, as far as that user is concerned. This can also be used to "uninstall" existing files (e.g. due to a renaming) - by letting "make install" install a file with Hidden=true in it. 	
boolean	NO	NO	1-4

FilePattern	A list of regular expressions to match against for a file manager to determine if this entry's icon should be displayed. Usually simply the name of the main executable and friends. 	
regexp(s)	NO	NO	1

TryExec	Filename of a binary on disk used to determine if the program is actually installed. If not, entry may not show in menus, etc. 	string	NO	NO	1
Exec	Program to execute, possibly with arguments. 	string	NO	YES	1
Path	If entry is type Application, the working directory to run the program in. 	
string	NO	YES	1

Terminal	Whether the program runs in a terminal window 	
boolean	NO	YES	1

SwallowTitle	If entry is swallowed onto the panel, this should be the title of window 	
localestring	NO	NO	1

SwallowExec	Program to exec if swallowed app is clicked. 	
string	NO	NO	1

Actions	Additional actions possible, see MIME type discussion in the section called “Detailed discussion of supporting MIME types”. 	
string(s)	NO	YES	1

MimeType	The MIME type(s) supported by this entry. 	
regexp(s)	NO	NO	1

SortOrder	This may specify the order in which to display files. 	
string(s)	NO	NO	4

Dev	The device to mount. 	
string	NO	NO	3

FSType	The type of filesystem to try to mount. 	
string	NO	NO	3

MountPoint	The mount point of the device in question. 	
string	NO	NO	3

ReadOnly	Specifies whether or not the device is read-only. 	
boolean	NO	NO	3

UnmountIcon	Icon to display when device is not mounted Mounted devices display icon from Icon key. UnmountIcons may be localized with the UnmountIcon[xx]= syntax. 	
string	NO	NO	3

URL	If entry is Link type, the URL to access. 	
string	NO	YES	2

Categories	Categories in which the entry should be shown in a menu (for possible values see the xdg-menu specification). 	
string(s)	NO	NO	1

OnlyShowIn / NotShowIn	A list of strings identifying the environments that should display/not display a given .desktop item. Only one of these keys, either OnlyShowIn or NotShowIn, may appear in a Group. (for possible values see the xdg-menu specification) 	
string(s)	NO	NO	1-4

StartupNotify	If true, it is KNOWN that the application will send a "remove" message when started with the DESKTOP_LAUNCH_ID environment variable set (see the startup notification spec for more details). 	
boolean	NO	NO	1

StartupWMClass	If true, it is KNOWN that the application will map at least one window with the given string as its WM class or WM name hint (see the startup notification spec for more details). 	
string	NO	NO	1

=cut
